/**
 * Smart water boiler system.
 * By Noa Lazar
 */
#include <WiFi.h>
#include <WebServer.h>
#include <OneWire.h>
#include <DallasTemperature.h>
#include "index.h"              // HTML interface page
#include "wifi_credentials.h"   // SSID and PASSWORD definitions
#include <ArduinoOTA.h>         // OTA update support
#include <time.h>               // Time handling for scheduling

WebServer server(80);           // Web server running on port 80

// Hardware pin definitions
#define RELAY_PIN 5
#define ONE_WIRE_BUS 4
#define LED_BLUE_BUILTIN 2

// Time configuration
const char* ntpServer = "pool.ntp.org";
const long gmtOffset_sec = 2 * 3600;
const int daylightOffset_sec = 3600;

// Control configuration
const float HYSTERESIS = 1.0;      // Temperature tolerance to avoid relay toggling too often
const long interval = 1000;        // Temperature check interval in milliseconds

// Global state variables
float currentTemp = 0;
float setPoint = 42.0;
bool relayState = false;
unsigned long previousMillis = 0;

// Schedule structure and variables
struct Schedule {
  int startHour;
  int startMinute;
  int endHour;
  int endMinute;
  float temp;
};

#define MAX_SCHEDULES 5
Schedule schedules[MAX_SCHEDULES];
int scheduleCount = 0;

// Control modes
enum ControlMode { MANUAL, TEMP_CONTROL, SCHEDULER };
ControlMode controlMode = TEMP_CONTROL;     // Selected mode by user
ControlMode activeMode = TEMP_CONTROL;      // Mode currently applied (depends on schedule and time)

// Temperature sensor setup
OneWire oneWire(ONE_WIRE_BUS);
DallasTemperature sensors(&oneWire);

// Determine core for temperature control task
#if CONFIG_FREERTOS_UNICORE
#define ARDUINO_RUNNING_CORE 0
#else
#define ARDUINO_RUNNING_CORE 1
#endif

void TaskControlTemperature(void *pvParameters);

/**
 * Returns the temperature setpoint based on current time and active schedules.
 */
float getScheduledTemperature(struct tm* timeinfo) {
  int nowMinutes = timeinfo->tm_hour * 60 + timeinfo->tm_min;
  for (int i = 0; i < scheduleCount; i++) {
    int start = schedules[i].startHour * 60 + schedules[i].startMinute;
    int end = schedules[i].endHour * 60 + schedules[i].endMinute;
    if (nowMinutes >= start && nowMinutes < end) {
      return schedules[i].temp;
    }
  }
  return -1;
}

// ==== Web Handlers ====

void handleRoot() {
  server.send(200, "text/html", MAIN_page);
}

/**
 * Returns current status as JSON for frontend requests.
 */
void handleServerMsg() {
  String json = "{";
  json += "\"temperature\":" + String(currentTemp, 1) + ",";
  json += "\"relay\":\"" + String(relayState ? "ON" : "OFF") + "\",";
  json += "\"mode\":\"" + String(controlMode == MANUAL ? "MANUAL" : 
                                 controlMode == TEMP_CONTROL ? "TEMP_CONTROL" : 
                                 "SCHEDULER") + "\",";
  json += "\"activeMode\":\"" + String(activeMode == MANUAL ? "MANUAL" : 
                                       activeMode == TEMP_CONTROL ? "TEMP_CONTROL" : 
                                       "SCHEDULER") + "\",";
  json += "\"schedulerActive\":" + String(scheduleCount > 0 ? "true" : "false");
  json += "}";
  server.send(200, "application/json", json);
}

void handleNotFound() {
  server.send(404, "text/plain", "404: Not Found");
}

/**
 * Task running on a separate core that handles temperature monitoring and relay control.
 */
void TaskControlTemperature(void *pvParameters) {
  for (;;) {
    unsigned long currentMillis = millis();
    if (currentMillis - previousMillis >= interval) {
      previousMillis = currentMillis;

      sensors.requestTemperatures();
      currentTemp = sensors.getTempCByIndex(0);

      if (controlMode == MANUAL) {
        activeMode = MANUAL;

      } else if (controlMode == TEMP_CONTROL) {
        activeMode = TEMP_CONTROL;

        if (!relayState && currentTemp < (setPoint - HYSTERESIS)) {
          relayState = true;
          digitalWrite(RELAY_PIN, LOW);
        } else if (relayState && currentTemp > (setPoint + HYSTERESIS)) {
          relayState = false;
          digitalWrite(RELAY_PIN, HIGH);
        }

      } else if (controlMode == SCHEDULER) {
        struct tm timeinfo;
        if (getLocalTime(&timeinfo)) {
          float scheduledTemp = getScheduledTemperature(&timeinfo);
          if (scheduledTemp > 0) {
            activeMode = SCHEDULER;
            setPoint = scheduledTemp;

            if (!relayState && currentTemp < (setPoint - HYSTERESIS)) {
              relayState = true;
              digitalWrite(RELAY_PIN, LOW);
            } else if (relayState && currentTemp > (setPoint + HYSTERESIS)) {
              relayState = false;
              digitalWrite(RELAY_PIN, HIGH);
            }
          } else {
            // No active schedule: disable relay
            activeMode = TEMP_CONTROL;
            relayState = false;
            digitalWrite(RELAY_PIN, HIGH);
          }
        }
      }

      Serial.printf("Temp: %.2f | ControlMode: %d | ActiveMode: %d | Relay: %s\n",
                    currentTemp,
                    controlMode,
                    activeMode,
                    relayState ? "ON" : "OFF");
    }

    vTaskDelay(10 / portTICK_PERIOD_MS);
  }
}

// ==== Setup and Loop ====

/**
 * Setup initializes WiFi, web server, OTA, sensors, and temperature task.
 */
void setup() {
  Serial.begin(115200);

  // Setup hardware
  pinMode(RELAY_PIN, OUTPUT);
  digitalWrite(RELAY_PIN, HIGH); // Relay OFF by default
  pinMode(LED_BLUE_BUILTIN, OUTPUT);
  digitalWrite(LED_BLUE_BUILTIN, LOW);
  delay(1000);
  sensors.begin();

  // Connect to WiFi
  WiFi.mode(WIFI_STA);
  WiFi.begin(SSID, PASSWORD);
  Serial.print("Connecting");

  while (WiFi.waitForConnectResult() != WL_CONNECTED) {
    Serial.print(".");
    digitalWrite(LED_BLUE_BUILTIN, !digitalRead(LED_BLUE_BUILTIN)); // Blink while connecting
    delay(250);
  }
  digitalWrite(LED_BLUE_BUILTIN, HIGH); // LED steady when connected
  Serial.println("\nConnected");
  Serial.print("IP Address: ");
  Serial.println(WiFi.localIP());

  configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);  // Set up NTP

  // Web routes
  server.on("/", handleRoot);
  server.on("/readTemperature", handleServerMsg);

  // Set target temperature manually
  server.on("/setTemperature", []() {
    if (server.hasArg("value")) {
      setPoint = server.arg("value").toFloat();
      controlMode = TEMP_CONTROL;
      server.send(200, "text/plain", "Target temperature set to " + String(setPoint));
    } else {
      server.send(400, "text/plain", "Missing value");
    }
  });

  // Manually control relay
  server.on("/relay", []() {
    if (server.hasArg("state")) {
      String state = server.arg("state");
      controlMode = MANUAL;
      if (state == "on") {
        relayState = true;
        digitalWrite(RELAY_PIN, LOW);
        server.send(200, "text/plain", "Relay manually turned ON");
      } else if (state == "off") {
        relayState = false;
        digitalWrite(RELAY_PIN, HIGH);
        server.send(200, "text/plain", "Relay manually turned OFF");
      } else {
        server.send(400, "text/plain", "Invalid state");
      }
    } else {
      server.send(400, "text/plain", "Missing state");
    }
  });

  // Change control mode
  server.on("/setMode", []() {
    if (server.hasArg("mode")) {
      String m = server.arg("mode");
      if (m == "MANUAL") controlMode = MANUAL;
      else if (m == "TEMP_CONTROL") controlMode = TEMP_CONTROL;
      else if (m == "SCHEDULER") controlMode = SCHEDULER;
      server.send(200, "text/plain", "Mode set to " + m);
    } else {
      server.send(400, "text/plain", "Missing mode");
    }
  });

  // Add a new schedule
  server.on("/setSchedule", []() {
    if (server.hasArg("start") && server.hasArg("end") && server.hasArg("temp")) {
      if (scheduleCount < MAX_SCHEDULES) {
        String start = server.arg("start");
        String end = server.arg("end");
        float temp = server.arg("temp").toFloat();
        int sh = start.substring(0, 2).toInt();
        int sm = start.substring(3).toInt();
        int eh = end.substring(0, 2).toInt();
        int em = end.substring(3).toInt();

        schedules[scheduleCount++] = { sh, sm, eh, em, temp };
        controlMode = SCHEDULER;
        server.send(200, "text/plain", "Schedule added");
      } else {
        server.send(400, "text/plain", "Schedule limit reached");
      }
    } else {
      server.send(400, "text/plain", "Missing parameters");
    }
  });

  // Clear all schedules
  server.on("/clearSchedule", []() {
    scheduleCount = 0;
    server.send(200, "text/plain", "All schedules cleared.");
  });

  server.onNotFound(handleNotFound);
  server.begin();

  // Setup OTA updates
  ArduinoOTA.setHostname("ESP32_Boiler");
  ArduinoOTA.setPassword(PASSWORD);
  ArduinoOTA.begin();

  // Launch temperature control task on specific core
  xTaskCreatePinnedToCore(
    TaskControlTemperature, "Temperature Control", 4096, NULL, 1, NULL, ARDUINO_RUNNING_CORE);
}

/**
 * Main loop handles OTA and client requests.
 */
void loop() {
  server.handleClient();
  ArduinoOTA.handle();
}
